/* daemon for the remote sensors
   cm@tahina.priv.at 2007
*/

#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <time.h>
#include <sys/errno.h>
#include <math.h>
#include <usb.h> /* written for libusb 0.1.12 */
#include "sensord.h"

struct usb_bus *busses;

struct {
  float v;
  float d1;
} d1_tab[] = {
  {10,   -41.25},			/* extrapolated */
  { 5,   -40},			/* from the data sheet */
  { 4,   -39.75},
  { 3.5, -39.66},
  { 3,   -39.6},
  { 2.5, -39.55},
  { 0,   -39.3}			/* extrapolated */
};

usb_dev_handle *find_sensor(void) {
  struct usb_bus *bus;
  char Manufacturer[STRINGSZ], Product[STRINGSZ];
  struct usb_device *dev;
  struct usb_dev_handle *devh = NULL;

  usb_init();
  usb_find_busses();
  usb_find_devices();
  busses = usb_get_busses();

  for(bus = busses; bus; bus = bus->next) {
    for(dev = bus->devices; dev; dev = dev->next) {
      if(dev->descriptor.idVendor == USB_VENDOR &&
	 dev->descriptor.idProduct == USB_PRODUCT) {
	devh = usb_open(dev);
	if(!devh) {
	  perror("usb_open");
	  goto xit;
	}
	if(usb_get_string_simple(devh, dev->descriptor.iManufacturer,
				 Manufacturer, STRINGSZ-1) < 0) {
	  perror("usb_get_string_simple(iManufacturer)");
	  goto xit;
	}
	if(usb_get_string_simple(devh, dev->descriptor.iProduct, Product,
				 STRINGSZ-1) < 0) {
	  perror("usb_get_string_simple(iProduct)");
	  goto xit;
	}
	printf("Found device \"%s\" \"%s\"\n", Manufacturer, Product);
	if(!strcmp(Manufacturer, USB_MANUF_STRING) &&
	   !strcmp(Product, USB_PRODUCT_STRING)) {
	  if(usb_set_configuration(devh, 1) < 0) {
	    perror("usb_set_configuration"); exit(1);
	  }
	  if(usb_claim_interface(devh, 0) < 0) {
	    perror("usb_claim_interface"); exit(1);
	  }
	  if((dev->config->interface->altsetting->endpoint->bmAttributes &
	      USB_ENDPOINT_TYPE_MASK) != USB_ENDPOINT_TYPE_INTERRUPT) {
	    fprintf(stderr, "Endpoint 1 is not an interrupt endpoint\n");
	    goto xit;
	  }
	  return devh;
	}
      }
    xit:
      if(devh) {
	usb_release_interface(devh, 0);
	usb_close(devh);
	devh = NULL;
      }
    }
  }
  return NULL;
}

void close_sensor(usb_dev_handle *devh) {
  if(devh) {
    usb_release_interface(devh, 0);
    usb_close(devh);
  }
}

double raw2temp(uint16_t v) {
  return -39.6 + 0.01 * v;
}

double raw2humid(uint16_t v) {
  return -4.0 + 0.0405 * v + -2.8e-6 * v * v;
}
  
void handle_packet(struct sensor_pkt *pkt, int len) {
  int i;
  time_t now;
  char tbuf[20];
  float temp, humid, volt, dewp, d1, H;
  FILE *cricket;

  if(len != sizeof(struct sensor_pkt)) {
    fprintf(stderr, "got packet len %d: ", len);
    for(i = 0; i < len; i++)
      fprintf(stderr, "%02x ", *(unsigned char *)(pkt+i));
    fprintf(stderr, "\n");
    return;
  }
  now = time(0);
  strftime(tbuf, 20, "%F %T", localtime(&now));

  volt = raw2voltage(pkt->vcc);

  /* interpolate d1 value for temperature */
  for(i = 0; d1_tab[i].v; i++) {
    if(volt > d1_tab[i].v) {
      if(i == 0) {
	d1 = d1_tab[i].d1;
      } else {
	d1 = (volt - d1_tab[i].v)/(d1_tab[i-1].v - d1_tab[i].v) *
	  (d1_tab[i-1].d1 - d1_tab[i].d1) + d1_tab[i].d1;
      }
    }
  }
  temp = d1 + 0.01 * pkt->temp;

  humid = -4 + 0.0405 * pkt->humid + -2.8e-6 * pkt->humid * pkt->humid;
  humid = humid + (temp - 25) * (0.01 + 0.00008 * pkt->humid);

  H = (log10f(humid)-2)/0.4343 + (17.62*temp)/(243.12+temp);
  dewp = 243.12*H/(17.62-H);

  printf("%s, %d, %d, %.2f, %.2f, %.2f, %.2f\n", tbuf, pkt->device_id,
	 pkt->sequence, temp, humid, volt, dewp);
  
  if(!(cricket = fopen(PATH_CRICKETF, "w"))) {
    perror("fopen " PATH_CRICKETF);
    exit(3);
  }
  fprintf(cricket, "# T\n%.2f\n# rH\n%.2f\n# Dewpoint\n%.2f\n# V\n%.2f\n",
	  temp, humid, dewp, volt);
  fclose(cricket);
}

  

int main(int argc, char **argv) {
  usb_dev_handle *devh;
  int len;
  char buffer[STRINGSZ];

  usb_set_debug(99);
  if(!(devh = find_sensor())) {
    fprintf(stderr, "No sensor found\n");
    exit(1);
  }

  while(1) {
    while(1) {
      len = usb_control_msg(devh, USB_TYPE_VENDOR | USB_RECIP_DEVICE |
			    USB_ENDPOINT_IN, VREQ_READ, 0, 0, buffer, 8,
			    1000);
      if(len < 0) {
	fprintf(stderr,"usb_control_msg: %s", usb_strerror());
	close_sensor(devh);
	exit(2);
      }
      if(len > 0) {
	//printf("VREQ_READ: %d bytes\n", len);
	handle_packet((struct sensor_pkt *)buffer, len);
      } else {
	break;
      }
    }


    //    if((len = usb_interrupt_read(devh, 1, buffer, 0, INT_MAX)) < 0) {
    if((len = usb_interrupt_read(devh, 1, buffer, 0, 10000)) < 0) {
      if(errno != EAGAIN) {
	fprintf(stderr, "usb_interrupt_read: %s\n", usb_strerror());
	close_sensor(devh);
	exit(2);
      }
    }
    // printf("interrupt_read: got %d\n", len);
  }

  close_sensor(devh);
}
    
