/* Name: receiver.c
 * Project: Remote Sensor; receiver based on AVR USB driver
 * Author: Christian Starkjohann
 * Creation Date: 2005-03-20
 * Tabsize: 4
 * Copyright: (c) 2005 by OBJECTIVE DEVELOPMENT Software GmbH
 * License: Proprietary, free under certain conditions. See Documentation.
 * This Revision: $Id: receiver.c 95 2006-01-26 18:38:21Z cs $
 */
#define F_CPU 12000000UL
#define BAUDRATE 5000

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include <util/delay.h>
#include <string.h>

#include "usbdrv.h"

#define LED_ON PORTD |= _BV(PD4)
#define LED_OFF PORTD &= ~(_BV(PD4))


static uchar    rxBuffer[12];
static uchar    rxIndex;

#define FIFO_SIZE   4   /* must be power of 2 */
#define FIFO_MASK   (FIFO_SIZE - 1)

static uchar    rxFifo[FIFO_SIZE][8];
static uchar    fifoWp, fifoRp, fifoIsFull;

struct _rcv_debug {
  uint16_t frame,
    badsof,
    crc,
    timeout,
    ok;
} rcv_debug;

/* ------------------------------------------------------------------------- */

static void fifoWrite(uchar *buffer)
{
uchar   i, *src, *dest;

    if(!fifoIsFull){
        src = buffer;
        dest = (uchar *)rxFifo + fifoWp;
        for(i=8;i;i--)
            *dest++ = *src++;
        fifoWp = (fifoWp + 8) & (FIFO_MASK << 3);
        fifoIsFull = fifoWp == fifoRp;
    }
}

static uchar    fifoSize(void)
{
uchar   size;

    if(fifoIsFull)
        return 4;
    /* enforce byte wide arithmetic: */
    size = (fifoWp - fifoRp);
    size >>= 3;
    size &= FIFO_MASK;
    return size;
}

static void fifoRead(uchar *buffer)
{
uchar   i, *src, *dest;

    if(fifoSize()){
        src = (uchar *)rxFifo + fifoRp;
        dest = buffer;
        for(i=8;i;i--)
            *dest++ = *src++;
        fifoRp = (fifoRp + 8) & (FIFO_MASK << 3);
        fifoIsFull = 0;
    }
    LED_OFF;
}

/* ------------------------------------------------------------------------- */
/* The receiver module returns random data (noise) if no carrier is received.
 * In order to ignore that babble, we do the following checks:
 *  - abort reception on byte framing errors
 *  - bytes must be received within tight time frame (40ms)
 *  - 16 bits start of frame signature must match
 *  - 4 bits value count must match
 *  - 16 bits CRC must match.
 * These checks provide at least 36 bits of security against false packets.
 */

static void usartTimeout(void)
{
    rxIndex = 0;
    rcv_debug.timeout++;
}

static void usartPoll(void)
{
uchar   c, e;
union{
    uchar       bytes[2];
    unsigned    word;
}       crc;

    e = UCSRA;  /* for error checking */
    c = UDR;
    rxBuffer[rxIndex++] = c;
    if(e & (1 << FE)){  /* framing error */
        rxIndex = 0;
	rcv_debug.frame++;
    }else{
        if((rxIndex == 1 && c != 0xdd) || (rxIndex == 2 && c != 0x16)){
            rxIndex = 0;    /* Start Of Frame error */
	    rcv_debug.badsof++;
        }else if(rxIndex == 1){     /* first byte of potential frame received */
            OCR1A = TCNT1 + 7500;   /* set up 40 ms timeout */
            TIFR = 1 << OCF1A;
        }else if(rxIndex >= 12){    /* complete frame received */
            crc.word = usbCrc16(rxBuffer, 10);
            if(crc.bytes[0] == rxBuffer[10] && crc.bytes[1] == c){  /* ignore if CRC bad, c == rxBuffer[11] here */
                uchar   x = rxBuffer[3] & 0xf0; /* force byte-wide operation */
                if(x == 0x30){  /* ignore if value count does not match */
                    fifoWrite(rxBuffer + 2);
                    usbSetInterrupt(0, 0);  /* NULL message on interrupt socket */
		    rcv_debug.ok++;
		    LED_ON;
                }
            } else {
	      rcv_debug.crc++;
	    }
            rxIndex = 0;
        }
	//	if(rxIndex > rcv_debug.ok) {
	//	  rcv_debug.ok = rxIndex;
	//	}
    }
}

/* ------------------------------------------------------------------------- */

static uchar    replyBuffer[10];
/* we don't need to store much status because we don't implement multiple
 * chunks in read/write transfers
 */

uchar   usbFunctionSetup(uchar data[8])
{
uchar   len = 0;

    if(data[1] == 0){       /* ECHO */
        len = 2;
        replyBuffer[0] = data[2];
        replyBuffer[1] = data[3];
	//replyBuffer[0] = 0xaa;
	//replyBuffer[1] = 0x55;
    }else if(data[1] == 1){ /* poll data available */
        len = 1;
        replyBuffer[0] = fifoSize();
    }else if(data[1] == 2){ /* read data */
        if(fifoSize()){
            len = 8;
            fifoRead(replyBuffer);
        }
    }else if(data[1] == 3) { /* get debug info */
      len = sizeof(rcv_debug);
      memcpy(replyBuffer, &rcv_debug, sizeof(rcv_debug));
    }
    
    usbMsgPtr = replyBuffer;
    return len;
}

/* These are for testing only -- we don't enable USB_CFG_IMPLEMENT_FN_* */
#if USB_CFG_IMPLEMENT_FN_WRITE
uchar   usbFunctionWrite(uchar *data, uchar len)
{
    return 0;
}
#endif
#if USB_CFG_IMPLEMENT_FN_READ
uchar   usbFunctionRead(uchar *data, uchar len)
{
    return 0;
}
#endif

/* ------------------------------------------------------------------------- */

int main(void)
{
    wdt_enable(WDTO_1S);
    PORTB = 0;			/* unused */
    DDRB = 0;

    PORTC = 0;			/* PC0 = ADC0 = RSSI */
    DDRC = 0;

    PORTD = _BV(PD3);		/* pullup for jumper */
    DDRD = _BV(PD4) | _BV(PD6);	/* LED, USB pullup */

    usbDeviceDisconnect();
    _delay_ms(15);		/* delay >10ms for USB reset */
    usbDeviceConnect();
    usbInit();
    sei();
    UCSRA = 0;
    UCSRB |= (1<<RXEN);
    UCSRC = _BV(URSEL) | _BV(UCSZ1) | _BV(UCSZ0); /* 8n1 */
    UBRRH = 0;
    UBRRL = F_CPU / (BAUDRATE * 16L) - 1;
    TCCR1B = 3; /* prescaler = 1/64 -> 1 count = 5.33333 us */
    for(;;){    /* main event loop */
        wdt_reset();
        usbPoll();
        if(UCSRA & (1 << RXC)){ /* USART RX complete */
            usartPoll();    /* must be polled at least every 6 ms */
        }
        if(TIFR & (1 << OCF1A)){
            TIFR = 1 << OCF1A;
            usartTimeout();
        }
    }
    return 0;
}

/* ------------------------------------------------------------------------- */
