/* remote temperature / humidity sensor */

#include "hardware.h"

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <util/delay.h>
#include <avr/eeprom.h>
#include <avr/sleep.h>
#include <avr/power.h>

#include "radioxmit.h"
#include "sht.h"

extern unsigned crc16(unsigned char *data, unsigned char len);

/* EEPROM addresses */
uint8_t EEMEM device_id_ee = 0;

/* globals */

uint8_t device_id;
uint8_t sequence;

struct {
  uint16_t sof;
  uint8_t devid;
  uint8_t seq;
  uint16_t data[3];
  uint16_t crc;
} buffer;

void read_device_id(void) {
  device_id = eeprom_read_byte(&device_id_ee);
}

/* do ADC conversion of Vbg vs Vcc */
uint16_t sample_vcc(void) {
  uint16_t res;

  power_adc_enable();
  set_sleep_mode(SLEEP_MODE_ADC);
  ADCSRA |= _BV(ADEN) | _BV(ADSC);
  while(ADCSRA & _BV(ADSC))
    sleep_mode();
  res = ADC;
  ADCSRA &= ~(_BV(ADEN));
  power_adc_disable();
  return res;
}

/* actually, nominal 64 seconds */
void wait_a_minute() {
  uint8_t i;
  for(i = 8; i; i--) {
    wdt_enable(WDTO_8S);
    WDTCR |= _BV(WDIE);
    set_sleep_mode(SLEEP_MODE_PWR_DOWN);
    sleep_mode();
  }
  wdt_disable();
}

void main(void) {
  int i;

  wdt_disable();
  
  /* data, enable to transmitter: */
  DDRB = _BV(PB1) | _BV(PB3);
  PORTB = _BV(PB0); /* test jumper */

  /* ADC: measure 1.1V internal reference against Vcc */
  /* prescaler /8 -> 125khz */
  /* single conv mode */
  ADMUX = _BV(MUX3) | _BV(MUX2);
  ADCSRA = _BV(ADIF) | _BV(ADIE) | /* clear interrupt flag, enable interrupt */
    _BV(ADPS1) | _BV(ADPS0);	/* prescaler /8 = 125kHz */
  ADCSRB = 0;

  power_all_disable();

  read_device_id();

  buffer.sof = 0x16dd;
  buffer.devid = device_id;
  buffer.seq = 0x30;

  sei();

  sht_init();

  while(1) {
    buffer.data[0] = sht_measure(SHT_MEAS_TEMP);
    buffer.data[1] = sht_measure(SHT_MEAS_HUMID);
    buffer.data[2] = sample_vcc();
    buffer.seq = 0x30 | (sequence++ & 0x0f);
    buffer.crc = crc16((unsigned char *)&buffer, 10);
    for(i = 3; i; i--) {
      xmit_pkt((unsigned char *)&buffer, sizeof(buffer));
      _delay_ms(5);
      }
    wait_a_minute();
  }
}

ISR(ADC_vect) { /* empty, but must exist for ADC conversion */
}

ISR(WDT_vect) { 
}
