/* routines to talk with sensirion.com's SHTxx temperature/humidity sensors
   cm@tahina.priv.at 2007
*/

#include "hardware.h"

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <avr/sleep.h>
#include <util/delay.h>

#include "sht.h"

//#define SHT_DEBUG

/* max SCK frequency <4.5V is 1MHz, so we don't need any wait states below
   that clock frequency */
#if F_CPU > 1000000UL
#error "Code doesn't support clock frequencies > 1MHz"
#endif

#define SHT_CONCAT(a, b)  a ## b
#define SHT_DDR(a) SHT_CONCAT(DDR, a)
#define SHT_PORT(a) SHT_CONCAT(PORT, a)
#define SHT_PIN(a) SHT_CONCAT(PIN, a)

#define DATA_IN() (SHT_PIN(SHT_DATA_PORT) & _BV(SHT_DATA_PIN))

#define DATA_DDR SHT_DDR(SHT_DATA_PORT)
#define DATA_PORT SHT_PORT(SHT_DATA_PORT)

#define SCK_DDR SHT_DDR(SHT_SCK_PORT)
#define SCK_PORT SHT_PORT(SHT_SCK_PORT)

#ifdef SHT_DEBUG
#define SCK_ON() (SCK_PORT |= _BV(SHT_SCK_PIN)); _delay_us(100);
#define SCK_OFF() (SCK_PORT &= ~(_BV(SHT_SCK_PIN))); _delay_us(100);
#else
#define SCK_ON() (SCK_PORT |= _BV(SHT_SCK_PIN)); __asm__ volatile("nop\n\tnop\n\t");
#define SCK_OFF() (SCK_PORT &= ~(_BV(SHT_SCK_PIN))); __asm__ volatile("nop\n\tnop\n\t");
#endif

#define XMIT_BIT(bit) set_data(bit); SCK_ON(); SCK_OFF()

static volatile uint8_t sht_int0_flag;

static inline void set_data(uint8_t bit) {
  if(bit) {
    DATA_DDR &= ~(_BV(SHT_DATA_PIN)); /* let the pullup signal the "1" */
  } else {
    DATA_DDR |= _BV(SHT_DATA_PIN);
  }
}

static inline uint8_t check_ack(void) {
  uint8_t error;
  set_data(1); /* float it */
  SCK_ON();
  error = DATA_IN();
  SCK_OFF();
  return !error;
}
  
  

/* transmission start sequence */
static void xmit_TS(void) {
  set_data(1);
  SCK_OFF();
  SCK_ON();
  set_data(0);
  SCK_OFF();
  SCK_ON();
  set_data(1);
  SCK_OFF();
}

static uint8_t xmit_byte(uint8_t byte) {
  int i;

  for(i = 8; i; i--) {
    XMIT_BIT(byte & 0x80);
    byte <<= 1;
  }
  if(!check_ack())
    return 0;
  else
    return 1;
}

static uint8_t xmit_cmd(uint8_t cmd) {
  xmit_TS();
  return xmit_byte(cmd);
}

uint16_t sht_measure(uint8_t what) {
  uint16_t value = 0;
  uint8_t bit;

  if(!xmit_cmd(what == SHT_MEAS_TEMP ? 3 : 5))
    return 0xfffe;		/* no ACK for CMD */
  set_data(1);			/* release line */

  /* SHT takes its time, we go to sleep in the meantime... */
  /* XXX need timeout */
  MCUCR &= ~(_BV(ISC01)|_BV(ISC00)); /* trigger on INT0 low level */
  sht_int0_flag = 0;
  GIMSK |= _BV(INT0);
  set_sleep_mode(SLEEP_MODE_IDLE); /* XXX PWR_DOWN?? */

  while(!sht_int0_flag)
    sleep_mode();

  if(DATA_IN()) {
    return 0xffff;
  }

  for(bit = 16; bit; bit--) {
    value <<= 1;
    SCK_ON();
    value |= (DATA_IN() ? 1 : 0);
    SCK_OFF();
    if(bit == 9) {		/* ACK first byte*/
      XMIT_BIT(0);
      set_data(1);
    } else if(bit == 1) {	/* NACK second byte, skip CRC */
      XMIT_BIT(1);
    }
  }
  return value;
}

/* sleep > 11ms after SHTxx initialisation */
static void sht_init_sleep(void) {
  wdt_enable(WDTO_30MS);	/* is the WDT accurate enough for WDT_15MS? */
  WDTCR |= _BV(WDIE);
  set_sleep_mode(SLEEP_MODE_PWR_DOWN);
  sleep_mode();
  wdt_disable();
}

uint8_t sht_init(void) {
  uint8_t i;

  DATA_PORT &= ~(_BV(SHT_DATA_PIN)); /* input, no pullup */
  DATA_DDR &= ~(_BV(SHT_DATA_PIN));
  SCK_PORT &= ~(_BV(SHT_SCK_PIN)); /* output, 0 */
  SCK_DDR |= _BV(SHT_SCK_PIN);

  /* wait 11ms after power on */
  sht_init_sleep();
  /* init interface, data high and 9 SCK pulses */
  set_data(1);
  for(i = 9; i; i--) {
    SCK_ON();
    SCK_OFF();
  }
  /* soft reset */
  xmit_cmd(0x1e); 		/* XXX error handling */
  sht_init_sleep();

  xmit_cmd(0x6);		/* write status register */
  /*  xmit_byte(0x2);	*/	/* no reload from OTP, 12/14 bit */
  xmit_byte(0);			/* defaults */

  return 1;
}

ISR(INT0_vect) {
  /* level triggered -- disable INT0 or we'll get back here as long as
     the level is low */
  GIMSK &= ~(_BV(INT0));
  sht_int0_flag = 1;
}
